import java.io.*;

class MyCache {
  private byte[] buf;
  int count;
  int browser;
  int platform;
  static int noOfFiles;
  FileInputStream cfile;
  static String[] locals;
  static String[] urls;
  String fname;

  final static int UNIX          = 21;
  final static int WIN           = 22;
  final static int EXPLORER      = 23;
  final static int NAVIGATOR     = 24;

  public MyCache(int plat, int brow) {
    int ptr;			// used as pointer while finding URLs
    buf = new byte[4096];	// record buffer for cache file URL records
    browser = brow;
    platform = plat;
    locals = new String[1000];	// local file names of cached files
    urls = new String[1000];	// urls of cached files

    // Open the cache file based on platform and browser type
    if (platform == UNIX)
      fname = "~/.netscape/cache/fat.db";
    else if (browser == NAVIGATOR) 
      fname = "\\NUNET\\NETSCAPE\\USERS\\MOS150\\CACHE\\FAT.DB";
    else fname = "\\WINDOWS\\TEMPORARY INTERNET FILES";
    try {
      cfile = new FileInputStream(fname);
    } catch( IOException e ){
      System.out.println("Error opening cache file!");
    }

    // Read cache file
    noOfFiles = 0;
    byte[] cb = new byte[512];	// read buffer (byte array)
    int i = 0;			// index to cb read buffer
    cb[0] = 0;
    int cnt = 1;
    count = 0;
    try {
      cnt  = cfile.read(cb);
      //System.out.println("Read " + cnt + " bytes");
      while (cnt > 0) {
        StringBuffer url = new StringBuffer("http:");
        while (cb[i] != 'h' && cb[i] != 'H' && i < cnt) {
          if (cb[i] != 'h' && cb[i] != 'H' && count == 4095) count = 0;
          buf[count++] = cb[i++];
        }
        if (i < cnt && (cb[i] == 'h' || cb[i] =='H')) {	// a URL might be found. Check coming characters
          int j;
          buf[count++] = cb[i++];
          //System.out.println("possible URL" + i + "cnt = " + cnt);
          for (j=1; j<=4 && i < cnt; j++) buf[count++] = cb[i++];
          if (j <= 4) {		// incomplete possible URL at the end of 512 byte block
            cnt  = cfile.read(cb);	// read next block
            //System.out.println("Read " + cnt + " bytes");
            if (cnt <= 0) break;
            i = 0;
            for (; j<=4 && i < cnt; j++) buf[count++] = cb[i++];	// copy remainder
          }

          if ((buf[count-4] == 't' && buf[count-3] == 't' &&
	      buf[count-2] == 'p' && buf[count-1] == ':') ||
              (buf[count-4] == 'T' && buf[count-3] == 'T' &&
	      buf[count-2] == 'P' && buf[count-1] == ':'))
          {
            // Found a URL. Now read the rest of it.
            //System.out.println("Found URL");
            ptr = count - 5;
            while (i < cnt && cb[i] > 0) url.append((char) cb[i++]);
            if (i == cnt && cb[i-1] != 0) {	// incomplete URL at the end of 512 byte block
              cnt  = cfile.read(cb);	// read next block
              //System.out.println("Read " + cnt + " bytes");
              if (cnt <= 0) break;
              i = 0;
              while (i < cnt && cb[i] > 0) url.append((char) cb[i++]);
            }

            // Now backtrack in the buffer (buf) and find the file type
            boolean found = false;
            int nonzero = 0;		// counts non-zero bytes backwards
            for (j=ptr-1; j>=0 && !found; j--) {
              if (buf[j] == 0) {
                if (nonzero > 4) { found = true; ptr = j; }
                nonzero = 0;
              }
              else nonzero++;
            }
            // Now backtrack in the buffer (buf) and find the local filename
            found = false;
            for (j=ptr-1; j>=0 && !found; j--) {
              if (buf[j] == 0) {
                if (nonzero > 4) { found = true; ptr = j; }
                nonzero = 0;
              }
              else nonzero++;
            }
            // Now copy the local filename into a variable
            if (found) {
              StringBuffer local = new StringBuffer((char) buf[++ptr]);
              while (buf[ptr] > 0) local.append((char) buf[ptr++]);
              System.out.println(local + " --> " + url);
              urls[noOfFiles] = url.toString();
              locals[noOfFiles] = local.toString();
              noOfFiles++;
            }

            count = 0;
          }		// end of "Found URL" if block
        }	// end of "possible URL" if block
        if (i == cnt) {		// are we at the end of a 512-byte block?
          cnt  = cfile.read(cb);	// if so, read the next block
          //System.out.println("Read " + cnt + " bytes");
          if (cnt <= 0) break;
          i = 0;
        }
      }  // while loop end of reading from cache file
    } catch(IOException e){
      System.out.println("Error reading cache file!");
    }
    try { cfile.close(); } catch (IOException e){
      System.out.println("Error closing cache file!");
    }
    System.out.println("No of Files: " + noOfFiles);
  }  // MyCache constructor

public static String localFile(String url) {
  for (int i=0; i<noOfFiles; i++)
    if (url.equals(urls[i])) return locals[i];
  return "";
}

  public static void main(String[] argv) {
    //System.out.println("Starting");
    MyCache mycache = new MyCache(WIN, NAVIGATOR);
    //MyCache mycache = new MyCache(WIN, EXPLORER);
    //System.out.println("Ending");
    String testurl = "http://www.gamelan.com/images/bro.gif";
    String testlocal = localFile(testurl);
    if (testlocal.equals("")) System.out.println("Local file of " + testurl + " is unknown");
    else System.out.println("Local file of " + testurl + " is " + testlocal);
    testurl = "test.com";
    testlocal = localFile(testurl);
    if (testlocal.equals("")) System.out.println("Local file of " + testurl + " is unknown");
    else System.out.println("Local file of " + testurl + " is " + testlocal);
  }

};