package webppserv;

import java.net.*;
import java.io.*;
import java.util.*;




///////////////////////////////////////////////////////////////////////////////
//
// Checksum Table
//
///////////////////////////////////////////////////////////////////////////////
class ChecksumTable extends Hashtable implements Serializable
{
    /* class variables */
    protected static boolean created = false;
    protected static boolean updated = false;
    protected static String serFileName = null;
    protected static File cTableFd;
    protected static String cTableDirName;    
    protected static ChecksumTable self = null;
    
    /* instance variables */
    protected boolean closed = false;
    
   /*
    *=========================================================================
    * Constructor
    *=========================================================================
    */    
    protected ChecksumTable()
    {
        super();
        /* intentionally left empty */
    }
    
   /*
    *=========================================================================
    * Replication Directory factory
    *=========================================================================
    */
    public static ChecksumTable factory()
    {
        File fd = null;
        ChecksumTable cTable = null;
        boolean status;
        String line = null;
        String thekey;
        String thevalue;
        int pos;
        
        if( created ){
            // checksum table is singleton 
            // Assert.assert( false );
            System.out.println( Webpp.webppRootDir +
                                ": detected another instance of checksum table" );                                
        }
         
        cTableDirName = Webpp.webppRootDir + 
                      System.getProperty( "file.separator" ) +
                      "replication";
        cTableFd = new File( cTableDirName );
      
        if( !cTableFd.exists() ){
            status = cTableFd.mkdirs();
            Assert.assert( status == true, 
                           "ChecksumTable.factory: could not create " + 
                           cTableDirName );
        }
      
        serFileName = cTableDirName + System.getProperty( "file.separator" ) +
                      "checksumTable.ser";
        fd = new File( serFileName );

        if( !fd.exists() ){
            // create a new checksum table
	      cTable = new ChecksumTable();
        } else {
            // deserialize an existing checksum table
		try {
              FileInputStream in = new FileInputStream( fd );
              ObjectInputStream oin = null;
		  if (in != null) {
		    oin = new ObjectInputStream( in );
		  }
		  if (oin != null) {
                cTable = (ChecksumTable) oin.readObject();
		    oin.close();
		  }
		  else {
		    System.out.println( "ChecksumTable deserialize error: oin is null");
		  }
		} catch (Exception e) {
		  System.out.println( "ChecksumTable deserialize exception: " + e.getMessage() );
		}
        }
                
        created = true;
	  updated = false;
        self    = cTable;
        return cTable;
    }

   /*
    *=========================================================================
    * serialize checksum table to the disk
    *=========================================================================
    */
    public void serialize()
    {
        FileOutputStream out = null;
        ObjectOutputStream oout = null;
        
        try {            
	    if (this.size() > 0) {
              out  = new FileOutputStream( serFileName );
	      if (out != null)
                oout = new ObjectOutputStream( out );
	    }
	    if (oout != null) {
              oout.writeObject( this );
              oout.flush();
              oout.close();
	    }
        } catch( IOException e ){
            System.out.println( "ChecksumTable serialize exception: " + e.getMessage() );
		return;
        }
	  updated = false;
    }


   /*
    *=========================================================================
    * close checksum table by serializing it on the disk
    *=========================================================================
    */
    public void close()
    {
	  if (updated) serialize();
	  closed = true;
    }
    
   /*
    *=========================================================================
    * serialize on the disk if not done
    *=========================================================================
    */
    protected void finalize()
    {
        if( !closed ){
            close();
        }
    }

   /*
    *=========================================================================
    * add a new item to the checksum table
    *=========================================================================
    */
    public synchronized Object put(Object key, Object value)
    {
	Object tmp = super.put(key, value);
	updated = true;
	//System.out.println("Checksum put: " + key + " " + value);
	return tmp;
    }

    public boolean updated()
    {
	return updated;
    }
}   
