package webppserv;

import java.io.*;
import java.lang.*;
import java.util.*;

class SegmentBuilder extends Thread implements Runnable
{
  int nextSegmentNo;
  static final int LocalSegmentHops = 5;	// # hops to consider a server in the same segment as the current server (itself)
  static final int MaxSegmentSize = 3;		// max # of servers in a segment
  static final int MinCommonHops = 4;		// min # common hops to place two other servers in the same segment

  public SegmentBuilder()
  {
    Webpp.traceroutesStarted = true;
    nextSegmentNo = 0;
  }

  public void run()
  {
        Enumeration enum;
        String serverName = null;
        String cmd = new String("traceroute");
        ThreadGroup tracerouteTG    = new ThreadGroup("Traceroute Group");

	yield();	// give main program a chance to go on

	System.out.println("Starting traceroutes...");

	Webpp.tracerouteCount = 0;
        enum = Webpp.serverPorts.keys();
        while( enum.hasMoreElements() ) {
          serverName = (String) enum.nextElement();
          if (serverName == null) return;
          if (serverName.equals(Webpp.server)) continue;  // don't call yourself
          Runnable r = new SysCommandThread(cmd, serverName);
          Thread t = new Thread(tracerouteTG, r);
          t.start();
          Webpp.tracerouteCount++;
        }
        while (Webpp.tracerouteCount > 0) {
	  try {
	    sleep(100);
	  } catch (InterruptedException ie) {}
	}
	if (Webpp.tracerouteDisabled) {
	  System.out.println("Traceroute does not work! Disabling it!");
	  return;
	}
        System.out.println("Traceroutes are finished");

	// Now, start the procedure to create the server segments
	enum = Webpp.traceroutePaths.keys();
	int arrSize = Webpp.traceroutePaths.size();
	int count = 0;
	String paths[] = new String[arrSize];
	int commons[] = new int[arrSize];
	int pathLengths[] = new int[arrSize];
	while( enum.hasMoreElements() ) {
	  serverName = (String) enum.nextElement();
	  if (serverName == null) return;
	  if (serverName.equals(Webpp.server)) continue;
	  paths[count++] = ((String) Webpp.traceroutePaths.get(serverName)) +
						" " + serverName;
	}

	// Sort the array of traceroute paths
	int i, j;
	String tmp;
	for (i=0; i<arrSize-1; i++)
	  for (j=i; j<arrSize-1; j++)
	    if (compareStr(paths[j], paths[j+1]) > 0) {
	      tmp = paths[j];
	      paths[j] = paths[j+1];
	      paths[j+1] = tmp;
	    }

	// First, determine the closest servers to myself and place
	// them into a segment with the current server itself.
	// Meanwhile, calculate the lengths of paths to each other server.
	// It also extracts the serverName for each traceroute path in the
	// 'paths' array.
	nextSegmentNo = 0;
	String serverNames[] = new String[arrSize];
	for (i=0; i<arrSize; i++) {
	  serverNames[i] = getServerName(paths[i]);
	  pathLengths[i] = getPathLength(paths[i]);
	  System.out.println(serverNames[i] + " # hops = " + pathLengths[i]);
	  if (pathLengths[i] <= LocalSegmentHops)
	    assignServers(Webpp.server, serverNames[i]);
	}

	// Calculate the similarities and differences between consecutive paths
	// Then, assign most similar ones to common segments gradually.
	String s1, s2;
	int maxCommon = 0;
	//boolean firstLoop = true;
	for (i=0; i < arrSize-1; i++) {
	  //s1 = serverNames[i];
	  //s2 = serverNames[i+1];
	  //if (Webpp.assigned.containsKey(s1) && Webpp.assigned.containsKey(s2))
          //  commons[i] = 0;
          //else
            commons[i] = commonStr(paths[i], paths[i+1]);
          if (commons[i] > maxCommon) maxCommon = commons[i];
        }
	while (maxCommon >= MinCommonHops) {
	//while (firstLoop || maxCommon >= MinCommonHops) {
	  //maxCommon = 0;
	  //firstLoop = false;

	  //if (maxCommon < MinCommonHops) break;
	  System.out.println("maxCommon=" + maxCommon);
	  for (i=0; i < arrSize-1; i++) {
	    if (commons[i] == maxCommon) {
	      s1 = serverNames[i];
              s2 = serverNames[i+1];
	      commons[i] = 0;
	      assignServers(s1, s2);	// assign server to a common segment
	    }
	  }
	  maxCommon = 0;
	  for (i=0; i < arrSize-1; i++)
	    if (commons[i] > maxCommon) maxCommon = commons[i];
	}

	// Assign unassigned servers to single-server segments by themselves.
	enum = Webpp.serverPorts.keys();
        while( enum.hasMoreElements() ) {
          serverName = (String) enum.nextElement();
          if (serverName == null) return;
	  assignServer(serverName);	// if unassigned, assign to a segment by itself
	}

	// Print out segment assignments (for debugging purpose)
	enum = Webpp.assigned.keys();
	while( enum.hasMoreElements() ) {
          serverName = (String) enum.nextElement();
          if (serverName == null) return;
	  Integer seg = (Integer) Webpp.assigned.get(serverName);
	  System.out.println(serverName + " assigned to " + seg);
	}

	Webpp.noOfSegments = nextSegmentNo;
	if (Webpp.respTimes != null)
	  Webpp.respTimes.setNoOfSegments(nextSegmentNo);
        Webpp.traceroutesDone = true;		// done
	Webpp.assigned.serialize();
  }

  // compareStr: compares two strings that contain traceroute paths.
  //	Return value is
  //		-1 if first string is smaller
  //		0  if strings are equivalent
  //		1  if first string is larger
  private int compareStr(String s1, String s2)
  {
    if (s1.equalsIgnoreCase(s2)) return 0;
    int i;
    int limit = s1.length();
    i = s2.length();
    if (i < limit) limit = i;		// min of two strings' lengths
    i = 0;
    for (i=0; i<limit && s1.charAt(i) == s2.charAt(i); i++);
    if (i >= limit) return 0;
    else if (s1.charAt(i) > s2.charAt(i)) return 1;
    else return -1;
  }

  // commonStr: compares two strings that contain traceroute paths.
  //	Return value is the number of network hops common to both paths.
  private int commonStr(String s1, String s2)
  {
    int common = 0, i;
    int limit = s1.length();
    i = s2.length();
    if (i < limit) limit = i;           // min of two strings' lengths
    i = 0;
    for (i=0; i<limit && s1.charAt(i) == s2.charAt(i); i++)
      if (s1.charAt(i) == ' ') common++;
    return common;
  }

  // getServerName: extracts the serveName from its traceroute path,
  //	and returns the serverName
  private  String getServerName(String line)
  {
    int pos = line.lastIndexOf(' ');
    if (pos < 0) return null;
    return line.substring(pos+1, line.length());
  }

  // getPathLength: returns the number of hops length of given traceroute
  // 	path.
  private int getPathLength(String line)
  {
    int i, hops = 0;
    if (line == null) return 0;
    for (i=0; i<line.length(); i++)
      if (line.charAt(i) == ' ')
	hops++;
    return hops;
  }

  // AssignServers: assigns two server to the same segment properly.
  // 	If one of the servers is already assigned to a segment, it adds the
  // 	other server to that segment.
  // 	If none of the servers is assigned to a segment yet, then creates
  // 	a new segment with those two servers.
  private void assignServers(String s1, String s2)
  {
    Integer seg, segSize;
    System.out.println("Assigning: " + s1 + " and " + s2);
    if (!Webpp.assigned.containsKey(s1) && !Webpp.assigned.containsKey(s2)) {
      seg = new Integer(nextSegmentNo++);
      Webpp.assigned.put(s1, seg);
      Webpp.assigned.put(s2, seg);
      segSize = new Integer(2);
      Webpp.segmentSizes.put(seg, segSize);
    }
    else if (Webpp.assigned.containsKey(s1) && !Webpp.assigned.containsKey(s2)) {
      seg = (Integer) Webpp.assigned.get(s1);
      segSize = (Integer) Webpp.segmentSizes.get(seg);
      if (segSize.intValue() < MaxSegmentSize)
	Webpp.assigned.put(s2, seg);
    }
    else if (!Webpp.assigned.containsKey(s1) && Webpp.assigned.containsKey(s2)) {
      seg = (Integer) Webpp.assigned.get(s2);
      segSize = (Integer) Webpp.segmentSizes.get(seg);
      if (segSize.intValue() < MaxSegmentSize)
	Webpp.assigned.put(s1, seg);
    }
    else {	// both s1 and s2 are already assigned to some segments
      seg = (Integer) Webpp.assigned.get(s1);
      Integer seg2 = (Integer) Webpp.assigned.get(s2);
      if (seg == seg2) return;
      segSize = (Integer) Webpp.segmentSizes.get(seg);
      Integer segSize2 = (Integer) Webpp.segmentSizes.get(seg2);
      if (segSize.intValue() + segSize2.intValue() > MaxSegmentSize) return;
      Integer theseg;
      String serverName;
      System.out.println("Joining segments " + seg + " and " + seg2);
      Enumeration enum = Webpp.assigned.keys();
      while( enum.hasMoreElements() ) {
        serverName = (String) enum.nextElement();
        if (serverName == null) return;
        theseg = (Integer) Webpp.assigned.get(serverName);
	if (theseg == seg2) Webpp.assigned.put(serverName, seg);
      }
    }
  }

  // AssignServer: assigns the given server to a new segment all by itself.
  //	This function is called to assigned lonely servers as single-server segments
  //	by themselves. It is called after all close enough servers are assigned
  //	together into segments.
  private void assignServer(String sName)
  {
    Integer seg, segSize;

    if (!Webpp.assigned.containsKey(sName)) {
      seg = new Integer(nextSegmentNo++);
      Webpp.assigned.put(sName, seg);
      segSize = new Integer(1);
      Webpp.segmentSizes.put(seg, segSize);
      System.out.println("Lonely server: " + sName);
    }
  }

}
