package webppserv;

import java.net.*;
import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import com.oroinc.text.regex.*;
import java.util.zip.*;		// Mehmet: added for checksum calculation
import java.lang.*;		// Mehmet: added for checksum calculation



///////////////////////////////////////////////////////////////////////////////
//
// Servlet implementing the Webpp functionality
//
///////////////////////////////////////////////////////////////////////////////
public class Webpp extends HttpServlet 
{
    /* class variables */
    static final int BUFFER_SIZE = 8192;   // buffer size used to read files
    // min no of request needed before making a replication decision
    static final long MinReqCountForDecision = 4;
    // min load percentage to make a replication decision
    static final double MinLoadPercentageForDecision = 0.35;
    // min load percentage out of totalRequestCount to consider a document as hot
    static final double MinReqPercentageForHotDoc = 0.35;
    // a clear for all counts is enforced when totalRequestCount reached the following number
    static final long LimitToEnforceClearCount = 100;

    static final int TRACEROUTE = 1;	// constant for this segmentation method
    static final int LATENCY = 2;		// constant for this segmentation method

    static long totalRequestCount = 0;	// total no of requests received for all resources
    static String rootDir = null;   // document root directory of the server
    static String webppRootDir;                      // webpp root directory
    static String server = null;                              // server name
    static int serverID = -1;			// Mehmet: server ID number
    static int port = -1;                              // server port number
    // Mehmet: clientHas string lists embedded resources of the requested URL
    // that are in client cache.
    String clientHas = null;
    
    static int tracerouteCount;	// keep track of # of threads for traceroute
    static int msgCount;	// keep track of # of threads for msg broadcasting
    static int segmentationMethod = LATENCY;	// how server segments are built (using TRACEROUTE or LATENCY)

    static boolean useCache = false;                    // if set use cache
    static boolean debugOutput = false;        // print output stream if set
    static boolean debugReplDir = false;     // debug msgs for repl dir ops
    static boolean debugCache = false;   // debug msgs related to cache use
    static boolean useChecksum = true;  // Mehmet: true -> replace emb res names with checksums
    static boolean clientHasParentDoc = false;	// Mehmet: will be set using request headers. If true, don't send parent doc.
    static boolean traceSuboperationTimes = false;	// Mehmet: measure times
    static boolean serversInitialized = false;	// to make sure we initialize other servers only once
    static boolean traceroutesDone;	// to make sure servlet creates its tree structure view and segments
    static boolean traceroutesStarted;	// to make sure we don't try to start traceroutes when they are in progress but but done yet
    static boolean tracerouteDisabled = false;	// set if a problem occurs with traceroute command so that the rest of programs do not use it
    static boolean latenciesStarted;	// to make sure we don't try to start latency measurement when it is in progress but but done yet
    static boolean latenciesDone;	// to make sure servlet measures its distance to all other servers

    // doNotSubstituteHyperlinks is normally set to false. It it set to true only
    //	The "DoNotSubstitute" header is received from the client, which means that
    //	this is not an ordinary client request but it is a request from another server
    //	by which the resources are downloaded onto that server (new replica or update).
    static boolean doNotSubstituteHyperlinks = false;

    /* Mehmet: Measure net checksum time for reading files and calculating
	checksums */
    static long netChecksumTime = 0;
    static Integer clientSegment = null;	// segment where the current request's initiator client belongs
    /* Mehmet: use a hash table to store calculated checksums during the
	service of the current request in order to avoid multiple checksum
	calculations for an embedded resource file */
    static ChecksumTable checksumHash;
    static Hashtable serverIDs;		// ID numbers of all servers
    static Hashtable serverPorts;	// port number used by each Web++ server
    static Hashtable serverStatus;	// is set to "Alive" or "Dead" for each server
    static Hashtable traceroutePaths;	// stores the traceroute paths to every other Web++ serevr
    static Hashtable HostServer;	// keeps track of which server is host for each client
    static SerHashtable assigned;	// tells which servers are assigned to which segments (Serializable Hash Table)
    static Hashtable segmentSizes;	// contains sizes of segments
    static ResponseTimeTable respTimes;	// response time info for each document and segment (server group)

    static int noOfSegments = 0;	// no of segments (server groups)
    static boolean BrowserNeedsData = false;	// Is set to true if servlet recognizes that browser cannot handle an empty response
    static String prevResponseTime = null;	// stores previous resp time piggybacked that is by client
    
    /* package variables */
    String reqFile = null; 
  
    /* private variables */
    protected static Cache cache = null;                     // cache object
    protected static ReplicationDir repl = null;   // replication dir object
    protected static Pattern parmSepPattern = null;  // param separator patt
    
    

   /*
    *=========================================================================
    * Constructor
    *=========================================================================
    */    
    public void init( ServletConfig config ) throws ServletException
    {
        super.init( config );

	traceroutesDone = false;	// Mehmet: to solve problem with Germany
	traceroutesStarted = false;	// Mehmet: to solve problem with Germany
	latenciesStarted = false;	// Mehmet: to solve problem with Germany
	latenciesDone = false;	// Mehmet: to solve problem with Germany

	System.out.println("In servlet init");
           
      webppRootDir = System.getProperty( "server.root" ) + 
                       System.getProperty("file.separator" ) + 
                       "webpp";
        
      cache = Cache.factory( this );
      repl  = ReplicationDir.factory();
	checksumHash = ChecksumTable.factory();
	assigned = SerHashtable.factory("segments.ser");
	
	readServerList();
                       
        try {
            parmSepPattern = SubstInputStream.compiler.compile( "&" );
        } catch( MalformedPatternException e ){
            throw new ServletException( "malformed parameter separator" );
        }
    }

         
   /*
    *=========================================================================
    * read documents with a substitution.
    * syntax: GET http://serverIP/servlet/webppserv.Webpp/<filename>
    *         (<filename> is relative to servlet home dir)
    *=========================================================================
    */    
    protected void doGet( HttpServletRequest req,
                          HttpServletResponse resp ) 
        throws ServletException, IOException
    {
        BufferedReader reader = null;
        ServletOutputStream out = null;
        BufferedInputStream in = null;
        String mimeType = null;
        byte buf[] = new byte[BUFFER_SIZE];
        int bytesRead;
        Date now;
        long t0 = 0, t1 = 0;	// Mehmet: Measure times

	//System.out.println("ReqType: " + req.getMethod());

	// Begin Mehmet: check query string first
	clientHas = req.getQueryString();
	String mainReq = req.getPathInfo();
	System.out.println("\nDocument: " + mainReq);
	//System.out.println("clientHas: " + clientHas);
	if (clientHas != null && ((clientHas.indexOf("ServerList") >=0) ||
		(clientHas.indexOf("HostServer") >=0) )) {
	  doPost(req, resp);
	  return;
	}
	// If there is info about previous request's resp time, extract it
	prevResponseTime = null;
	if (clientHas != null) {
	  int pos1 = clientHas.indexOf('|');
	  if (pos1 >= 0) {
	    int pos2 = clientHas.lastIndexOf('|');
	    if (pos2 > pos1) {
	      prevResponseTime = clientHas.substring(pos1+1, pos2);
	      System.out.println("prevResponseTime: " + prevResponseTime);
	    }
	    else {
	      System.out.println("Error: cannot extract prev response time properly!");
	    }
	    if (pos1 == 0) clientHas = null;
	    else clientHas = clientHas.substring(0, pos1);
	    //System.out.println("New clientHas: " + clientHas);
	  }
	}
	// End Mehmet

	// Begin Mehmet: find the client's segment if you can.
	String clientAddr = req.getRemoteAddr();
	clientSegment = getSegmentofClient(clientAddr);
	if (clientSegment != null)
	  System.out.println("Segment of client is " + clientSegment);
	// End Mehmet

	// Begin Mehmet: if prevRespTime and clientSegment could be determined,
	//	then insert the information into the respTimes hash table.
	if (clientSegment != null && prevResponseTime != null)
	  insertRespTime(clientSegment, prevResponseTime);
	// End Mehmet

	// Begin Mehmet: debug headers
	//Enumeration reqHeaders = req.getHeaderNames();
	//while (reqHeaders != null && reqHeaders.hasMoreElements()) {
	//  String headerStr = ((String) reqHeaders.nextElement());
	//  String valueStr = null;
	//  if (headerStr != null) 
	//    valueStr = req.getHeader(headerStr);
	//  if (headerStr != null && valueStr != null)
	//    System.out.println("Header: " + headerStr + " : " + valueStr);
	//}
	// End Mehmet: debug headers
	
	if (traceSuboperationTimes)
	  t0 = System.currentTimeMillis();	// Mehmet: Measure times

	netChecksumTime = 0;	// Mehmet: Measure times

        if( server == null ){
            server = req.getServerName();
        }
        if( port == -1 ){
            port = req.getServerPort();
        }
        if( rootDir == null ){
            rootDir = req.getRealPath( "/" );
        }
        
	// Mehmet: get server ID number
	Integer tmpid = new Integer((String) serverIDs.get(server));
	serverID = tmpid.intValue();
	System.out.println("Server-" + serverID + " at " + server + " is up");
	
	// Mehmet: initialize all servers so that they create server clusters
	if (serverID == 0 && !serversInitialized) {
	  initServers();
	  if (segmentationMethod == LATENCY && !latenciesStarted && !latenciesDone)
	    doLatencies();

	}

	// Mehmet: do traceroutes to create tree structure view of network
      if (segmentationMethod == TRACEROUTE && !traceroutesStarted && !traceroutesDone)
	  doTraceroutes();
	          
      reqFile = req.getPathTranslated();
      if( reqFile == null ){
        throw new ServletException( "no extra path or could not xlate" );
      }

      // Begin Mehmet
	//System.out.println("Mehmet: reqFile:" + reqFile);	// Mehmet: debug
	//System.out.println("Document: " + req.getPathInfo());		// Mehmet: debug
	//System.out.println("Client has: " + clientHas);	// Mehmet: debug
	String tmp = req.getHeader("DoNotSubstitute");
	if (tmp != null && tmp.equalsIgnoreCase("Yes"))
	  doNotSubstituteHyperlinks = true;
	else doNotSubstituteHyperlinks = false;
	tmp = req.getHeader("ClientHasParentDoc");
	if (tmp == null) {	// The browser might be Netscape, so check clientHas
	  //System.out.println("Header NOT received");
	  if (clientHas != null && clientHas.indexOf(";parent;") != -1) {
	    BrowserNeedsData = true;
	    tmp = "yes";
	  }
	}
	else {
	  //System.out.println("Header received");
	  BrowserNeedsData = false;	// Browser is not Netscape, so it can handle empty response properly
	}
	if (tmp != null && tmp.equalsIgnoreCase("yes")) {
	  //System.out.println("Client has parent doc");	// Mehmet: debug
	  clientHasParentDoc = true;
	}
	else {
	  //System.out.println("Client does not have parent doc");	// Mehmet: debug
	  clientHasParentDoc = false;
	}
	// End Mehmet
        
        mimeType = getMimeType( reqFile );
        resp.setContentType( mimeType );
	//resp.setHeader("TestHeader","nanik");		// Mehmet: debug headers

	  // Begin Mehmet: replaced the following if-else block with the
	  //	resp.setDateHeader call below. I do not want the indexWebpp.html
	  //	to be cached by the client any more.
	  resp.setDateHeader( "Expires", 0 );
/*
        if( reqFile.indexOf("indexWebpp.html") != -1 ){
            // set Expires for root document to 30 min from now
            //     allows browsers to return to parent docs even when 
            //     the server is down
            now = new Date();
            resp.setDateHeader( "Expires", now.getTime() + 1800000 );
	    // Begin Mehmet: get the list of server names and send them to the
	    //	client.
	    //String serverNamesStr = getActiveServerNamesList();
	    //if (serverNamesStr != null)
	    //  resp.setHeader("ServerList", serverNamesStr);	// Mehmet send the list of server names with port numbers
	    // End Mehmet
        } else {
            // all other docs are non-cacheable
            resp.setDateHeader( "Expires", 0 );
        }
*/
        // End Mehmet

        out = resp.getOutputStream();

	// Begin Mehmet: check if client has all emb resources and parent doc.
	// If so, then skip the rest of the operations so that client can
	// immediately its cached copies of resources.
	if (clientHasParentDoc && clientHas.equalsIgnoreCase("ALL")) {
	    //System.out.println("Client has ALL, so servlet does not send anything");	// Mehmet: debug
	    return;
	}
	// End Mehmet

	// Begin Mehmet: if client cache contains the parent document, check if it also
	//	contains all embedded resources. If it does, then simply skip the rest
	//	of the procedures and return nothing to the client. Client will use the
	//	local copies automatically.
	String fileSep = System.getProperty( "file.separator" );
	String parentDocDir = reqFile.substring( 0, 
        			reqFile.lastIndexOf(fileSep)+1 );

	if (clientHasParentDoc) {
	  String EmbResList = ReplicationDir.getEmbeddedResources( reqFile );
	  boolean hasAll = true;
	  if (EmbResList != null) {
	    StringTokenizer st = new StringTokenizer(EmbResList, ";");
	    for ( ; hasAll && st.hasMoreElements(); ) {
	      if (useChecksum)
		tmp = getChecksumValue(new Adler32(), parentDocDir +
					((String) st.nextElement()));
	      else
		tmp = ";" + ((String) st.nextElement()) + ";";
	      if (clientHas.indexOf(tmp) == -1) hasAll = false;
	    }
	  }
	  if (hasAll) {
	    //System.out.println("Client has all resources and parent doc");
	    if (BrowserNeedsData) {	// Browser cannot handle an empty response (most probably Netscape)
	      out.print("UseLocalCopy");	// Don't send empty response, just send something
	      out.close();
	    }
	    return;
	  }
	}
	// End Mehmet
	
	if (traceSuboperationTimes) {
	  t1 = System.currentTimeMillis() - t0;	// Mehmet: Measure times
	  System.out.println("\nProcessing Headers: " + t1);	// Mehmet: Measure times
	  t0 = System.currentTimeMillis();	// Mehmet: Measure times for next subop
	}
        try {
	    if (doNotSubstituteHyperlinks)
	      in = new BatchInputStream(reqFile, this);
	    else {
		if (useCache)
              in = cache.getFile( reqFile );	// Mehmet: inserted if-else to this line
		else
		  in = new SubstInputStream( reqFile, this );
	    }
        } catch( FileNotFoundException e ){
            resp.sendError( HttpServletResponse.SC_NOT_FOUND, 
                            "Could not find file " + reqFile );
            return;
        }
	if (traceSuboperationTimes) {
	  t1 = System.currentTimeMillis() - t0;	// Mehmet: Measure times
	  System.out.println("Preparing batch file: " + t1);	// Mehmet: Measure times
	  t0 = System.currentTimeMillis();	// Mehmet: Measure times for next subop
	}

	while( (bytesRead = in.read(buf, 0, buf.length)) != -1 ){
            out.write( buf, 0, bytesRead );
            
            if( Webpp.debugOutput ){
                System.out.write( buf, 0, bytesRead );
            }
        }

	if (traceSuboperationTimes) {		// Mehmet: Measure times
	  t1 = System.currentTimeMillis() - t0;	// Mehmet: Measure times
	  System.out.println("Sending batch file: " + t1);	// Mehmet: Measure times
	}

        try {
            in.close();
        } catch( IOException e ){
            Assert.assert( false );
        }

	// Begin Mehmet
	// If the checksum table is updated, then serialize it to disk
	if (checksumHash.updated())
	  checksumHash.serialize();
	// End Mehmet

    }


   /*
    *=========================================================================
    * Update of a document, no update of replication directory
    * Stx: PUT http://serverIP/servlet/webppserv.Webpp?<urn>&<val2>&..
    * FIXME: Currently, <urn> serves both as a logical URl and as a local
    *        filename. Later we will need to decouple those two.
    *=========================================================================
    */
    protected void doPut( HttpServletRequest req,
                          HttpServletResponse resp )
        throws ServletException, IOException
    {
        ServletInputStream reader = null;
        String storeFile = null;
        FileOutputStream store = null;
        String line = null;
        String qry = null;
        Vector pars;
        String urn = null;
        String urlList = null; 
        byte buffer[] = new byte[4096];
        int len = 0;

        // get qry string
	    qry = req.getQueryString();
	    if( qry == null ){
	        resp.sendError( HttpServletResponse.SC_NOT_FOUND,
	                        "no query string received" );
	        return;
	    }

	// parse qry string
	    pars = Util.split( SubstInputStream.matcher, parmSepPattern, qry );
            for( int i = 0; i < pars.size(); i++ ){
            if( i == 0 ){
                urn = (String) pars.elementAt( i );
            } else if( i == 1 ){
                urlList = (String) pars.elementAt( i );
            } else {
                urlList += " " + (String) pars.elementAt( i );
            }
        }

        // open the stream to read the file to replicate
        reader = req.getInputStream();

        // open the stream to store the file
        storeFile = req.getRealPath( urn );

        if( storeFile == null ){
            throw new ServletException( "could not xlate path " + urn );
        }
        try {
            store  = new FileOutputStream( storeFile );

        } catch( IOException e ){
            resp.sendError( HttpServletResponse.SC_NOT_FOUND,
                            "could not open store file " + storeFile );
        }

        // store the file
        while( (len = reader.read( buffer )) != -1 ){
            store.write( buffer, 0, len );
        }
        store.close();
       
        resp.setStatus( HttpServletResponse.SC_OK,
                        "replicated file " + storeFile );
    }

        
   /*
    *=========================================================================
    * Mehmet: Inter-servlet communication is carried on by using this function.
    * Stx: POST http://serverIP/servlet/webppserv.Webpp?request&key1=val1&key2=val2...
    * 'request' gives us the request type, key and value pairs contain the parameters.
    *=========================================================================
    */
    protected void doPost( HttpServletRequest req,
                           HttpServletResponse resp )
        throws ServletException, IOException
    {
	ServletOutputStream out = null;
	String qry = null, reqType = null, parameters = null;
        Vector pars;
	int len = 0;
	byte buffer[] = new byte[4096];

	//System.out.println("ReqType: " + req.getMethod());

	// get qry string
	qry = req.getQueryString();
	if( qry == null ) {
	  resp.sendError( HttpServletResponse.SC_NOT_FOUND,
					"doPost: no query string received" );
	  return;
	}

	System.out.println("Post query str: " + qry);
	
	// parse qry string
	pars = Util.split( SubstInputStream.matcher, parmSepPattern, qry );
	for( int i = 0; i < pars.size(); i++ ){
	  if( i == 0 ){
	    reqType = (String) pars.elementAt( i );
	  } else if( i == 1 ){
	    parameters = (String) pars.elementAt( i );
	  } else {
	    parameters += " " + (String) pars.elementAt( i );
	  }
	}

	if (reqType == null) {
	  resp.sendError( HttpServletResponse.SC_NOT_FOUND,
					"doPost: no request type specified");
	  return;
	}

	out = resp.getOutputStream();

	if (reqType.equalsIgnoreCase("ServerList")) {
	  // Read, but ignore the contents of input stream for this req Type.
	  // reader = req.getInputStream();
	  //while( (len = reader.read( buffer )) != -1 );
	  String serverNamesStr = getActiveServerNamesList();
	  if (serverNamesStr != null) {
	    //resp.setHeader("ServerList", serverNamesStr);	// Mehmet send the list of server names with port numbers
	    out.print(serverNamesStr);		// Mehmet send the list of server names with port numbers
	  }
	}	// ServerList
	else if (reqType.equalsIgnoreCase("HostServer")) {
	  String clientAddr = req.getRemoteAddr();
	  if (clientAddr != null) {
	    HostServer.put(clientAddr, server);		// put the client into your table
	    System.out.println("Received msg: " + clientAddr + " says I am its host");
	    broadcastMessage("HostServerBroadcast", clientAddr + " " + server);	// Tell other servers you're this client's host
	    System.out.println("Broadcasted msg: I am host for " + clientAddr);
	    out.print("Success");
	  }
	}	// HostServer
	else if (reqType.equalsIgnoreCase("HostServerBroadcast")) {
	  String msg = req.getHeader("param");
	  if (msg != null) {
	    StringTokenizer st = new StringTokenizer(msg, " ");
	    if (st.hasMoreElements()) {
	      String clientAddr = (String) st.nextElement();
	      String serverName = null;
	      if (st.hasMoreElements()) {
		serverName = (String) st.nextElement();
		HostServer.put(clientAddr, serverName);		// insert to table
	      
		System.out.println("Received msg: Host of " + clientAddr +
						" is " + serverName);
		out.print("Success");
	      }
	      else {
		out.print("Failed");
	      }
	    }
	    else {
	      out.print("Failed");
	    }
	  }
	  else {
	    out.print("Failed");
	  }
	}	// HostServerBroadcast
	else if (reqType.equalsIgnoreCase("AliveServers")) {
	  String msg = req.getHeader("param");
	  if (msg != null) {
	    System.out.println("Alive servers received: " + msg);
	    StringTokenizer st = new StringTokenizer(msg, " ");
	    while (st.hasMoreElements()) {
	      String serverName = (String) st.nextElement();
	      serverStatus.put(serverName, "Alive");
	    }
	    if (segmentationMethod == LATENCY && !latenciesStarted && !latenciesDone)
		doLatencies();
	  }
	}	// AliveServers
	else if (reqType.equalsIgnoreCase("ReplicationRequest")) {
	  String msg = req.getHeader("param");
	  if (msg != null) {
	    String requestingServer = req.getRemoteHost();
	    if (requestingServer != null) requestingServer = requestingServer.toLowerCase();
	    System.out.println("Replication request: " + requestingServer + " wants to replicate " +
			msg + " on this server");
	    int pos = msg.indexOf(' ');
	    if (pos == -1) {
	      System.out.println("Error: cannot parse filename and url in replication request msg");
	      out.print("Failed");
	    }
	    else {	// msg successfully parsed
	      String fname = msg.substring(0, pos);
	      String url = msg.substring(pos + 1);
	      if (serverContainsResource(server, fname)) {
		System.out.println("Error: I already have a replica of " + fname);
		out.print("Failed");
	      }
	      else {	// it is okay to do the replication
		downloadResource(requestingServer, fname, url);
		out.print("Success");
	      }
	    }
	  }
	}	// ReplicationRequest
	else if (reqType.equalsIgnoreCase("AddToReplDir")) {
	  String msg = req.getHeader("param");
	  if (msg != null) {
	    String requestingServer = req.getRemoteHost();
	    System.out.println("AddToReplDir request received: " + requestingServer + " adds " +
			msg + " to replication directory");
	    int pos = msg.indexOf(' ');
	    if (pos == -1) {
	      System.out.println("Error: cannot parse filename and url in AddToReplDir request msg");
	      out.print("Failed");
	    }
	    else {	// msg successfully parsed
	      String fname = msg.substring(0, pos);
	      String url = msg.substring(pos + 1);
	      repl.appendUrl(fname, url);
	      out.print("Success");
	    }
	  }
	}	// AddToReplDir

	out.close();
    }
    
    
   /*
    *=========================================================================
    * Destructor
    *=========================================================================
    */    
    public void destroy()
    {
        cache.close();
        repl.close();
	checksumHash.close();
    }
    
    
   /*
    *=========================================================================
    * Deduce mime type from file suffix (the Apache way)
    *=========================================================================
    */        
    protected String getMimeType( String fileName )
    {
        String mimeType;
        
        mimeType = getServletConfig().getServletContext().
                   getMimeType( fileName );
                   
        if( mimeType == null ){
            return "text/plain";
        }
        
        return mimeType;
        
        /*
        dotIx = fileName.lastIndexOf( '.' );
        
        if( dotIx == -1 ){
            return "text/plain";
        }
        
        s = fileName.substring( dotIx+1 );
        */
        /********************************************************************
         FIXME: at the time the list is not complete, more types should be
                put in for a working prototype.
         ********************************************************************/
        /*
        if( s.equals("html") || s.equals("htm") ){
            return "text/html";
        }
        
        if( s.equals("txt") ){
            return "text/plain";
        }
        
        if( s.equals("mpeg") || s.equals("mpg") || s.equals("mpe") ){
            return "video/mpeg";
        }
        
        if( s.equals("gif") ){
            return "image/gif";
        }
        
        if( s.equals("jpeg") || s.equals("jpg") || s.equals("jpe") ){
            return "image/jpeg";
        }
        
        if( s.equals("png") ){
            return "image/png";
        }
        
        if( s.equals("tiff") || s.equals("tif") ){
            return "image/tiff";
        }
        
        if( s.equals("ai") || s.equals("eps") || s.equals("ps") ){
            return "application/postscript";
        }  
        
        if( s.equals("ppt") ){
            return "application/powerpoint";
        }
        
        if( s.equals("pdf") ){
            return "application/pdf";
        }
        
        if( s.equals("doc") ){
            return "application/msword";
        }
        
        if( s.equals("zip") ){
            return "application/zip";
        }
        
        return "text/plain";
        */
    }

   /*
    *=========================================================================
    * getChecksumValue: Calculates the checksum for the file whose name is given as argument.
    * Mehmet: this fcn is called while parsing the parent HTML document for embedded resources.
    *=========================================================================
   */

    static String getChecksumValue(Checksum checksum, String fname) {
	String result = (String) checksumHash.get(fname);	// is this checksum already calculated during this request's service
	if (result != null) {
	  //System.out.println("Checksum found in hash table for " + fname);
	  return result;	// return the checksum that was already calculated and stored in hash table
	}
        //System.out.println("Calculating checksum for " + fname);
        long t0 = System.currentTimeMillis();		// Mehmet: Measure times
        try {
            BufferedInputStream is = new BufferedInputStream(
                new FileInputStream(fname));
            byte[] bytes = new byte[1024];
            int len = 0;
            
            // Read the file and compute the checksum.
            while ((len = is.read(bytes)) >= 0) {
                checksum.update(bytes, 0, len);
            }
            is.close();
        } catch (IOException e) {
            //e.printStackTrace();
            netChecksumTime += System.currentTimeMillis() - t0;	// Mehmet: Measure times
	    return null;
        }
        long chksm = checksum.getValue();
	result = new String();
	result = result.valueOf(chksm);
	netChecksumTime += System.currentTimeMillis() - t0;	// Mehmet: Measure times
	checksumHash.put(fname, result);
	return result;
    }


   /*
    *=========================================================================
    * serverIsAlive: returns true is that server is alive, false otherwise
    *=========================================================================
    */
    static boolean serverIsAlive(String serverName)
    {
      if (serverName.equals(server)) return true;
      String serverStat = (String) serverStatus.get(serverName);
      if (serverStat != null && serverStat.equalsIgnoreCase("Alive")) return true;
      else return false;
    }


   /*
    *=========================================================================
    * getActiveServerNamesList: collects active Web++ server names into one string object
    *=========================================================================
    */
    static String getActiveServerNamesList()
    {
      Enumeration enum = serverPorts.keys();
      String serverName = null, serverNamesStr = null;
      while( enum.hasMoreElements() ) {
	  serverName = (String) enum.nextElement();
	  if (!serverIsAlive(serverName)) continue;	// skip if that server is dead
	  serverName += ":" + ((String) serverPorts.get(serverName));
	  if (serverNamesStr == null)
	    serverNamesStr = new String(serverName);
	  else
	    serverNamesStr += " " + serverName;
	}
	return serverNamesStr;
    }


   /*
    *=========================================================================
    * getServerNamesList: collects all Web++ server names into one string object
    *=========================================================================
    */
    static String getServerNamesList()
    {
      Enumeration enum = serverPorts.keys();
      String serverName = null, serverNamesStr = null;
      while( enum.hasMoreElements() ) {
	serverName = (String) enum.nextElement();
	serverName += ":" + ((String) serverPorts.get(serverName));
	if (serverNamesStr == null)
	  serverNamesStr = new String(serverName);
	else
	  serverNamesStr += " " + serverName;
	}
	return serverNamesStr;
    }


   /*
    *=========================================================================
    * readServerList: reads list of servers and info about each server.
    *   The information is inserted into serverIDs and serverPorts hash
    *   tables.
    *=========================================================================
    */

    static void readServerList()
    {
	BufferedReader in = null;
	File fd = null;
	String line = null, str1, str2;
	int pos1, pos2;
	String dirName =  Webpp.webppRootDir +
                      System.getProperty( "file.separator" ) +
                      "replication";
	File dirFd = new File( dirName );

	if( !dirFd.exists() )
	  Assert.assert( true,"Cannot find directory: " + dirName );
      
	String serFileName = dirName + System.getProperty( "file.separator" ) +
							"servers.ser";
	fd = new File( serFileName );
	serverIDs = new Hashtable();
	serverPorts = new Hashtable();
	serverStatus = new Hashtable();
	traceroutePaths = new Hashtable();
	HostServer = new Hashtable();
	segmentSizes = new Hashtable();
	if (assigned != null && !assigned.isNew()) {
	  int segCount = getNoOfSegmentsFromTable();
	  noOfSegments = segCount;
	}
	respTimes = ResponseTimeTable.factory(noOfSegments);
			
	if( fd.exists() ){
	  try {
            in = new BufferedReader( new FileReader( fd ) );
          } catch( Exception e ){
            Assert.assert( false, e.getMessage() );
          }
          try {
            while( (line = in.readLine()) != null ){
              pos1 = line.indexOf( ' ' );
              if( pos1 == -1 ) continue;
              str1 = line.substring( 0, pos1 );
              pos2 = line.indexOf( ' ', pos1+1);
              if( pos2 == -1 ) continue;
              str2 = line.substring( pos1+1, pos2 );
              serverIDs.put(str1, str2);
              pos1 = pos2;
              pos2 = line.indexOf( ' ', pos1+1);
              if( pos2 == -1 ) pos2 = line.length();
              str2 = line.substring( pos1+1, pos2 );
              serverPorts.put(str1, str2);
            }
          } catch( Exception e ){
	    Assert.assert( false, e.getMessage() );
	  }
	}
	else {
	  Assert.assert( true,"Cannot find file: " + serFileName );
        }
    }
    
    /*
    *=========================================================================
    * initServers: Sends dummy requests to other servers so that they start
    *   constructing the server clusters. Then, it calls the function at the
    *   local server to start the same operation.
    *=========================================================================
    */

    static void initServers()
    {
	Enumeration enum;
	String serverName = null, serverNames = server;
	GetURL myGetURL = new GetURL();
	long bytesRead;
	
	serverStatus.put(server, "Alive");	// put my status as Alive

	enum = serverPorts.keys();
	while( enum.hasMoreElements() ) {
	  bytesRead = 0;
	  serverName = (String) enum.nextElement();
	  if (serverName == null) return;
	  if (serverName.equals(server)) continue;	// don't call yourself
	  try {
	    bytesRead = myGetURL.getURL("http://" + serverName + ":" +
	    		serverPorts.get(serverName) + 
	    		"/servlet/webppserv.Webpp/indexWebpp.html");
	  } catch (Exception e) {
	    System.out.println("Error while calling server " + serverName);
	    serverStatus.put(serverName, "Dead");
	    bytesRead = 0;
	  }
	  if (bytesRead > 0) {
	    System.out.println(serverName + " is contacted successfully");
	    serverStatus.put(serverName, "Alive");
	    if (serverNames == null) serverNames = serverName;
	    else serverNames += " " + serverName;
	  }
	}
	broadcastMessage("AliveServers", serverNames);
	serversInitialized = true;
    }

    /*
    *=========================================================================
    * doTraceroutes: Sends traceroute requests to other servers in order to
    *   construct the tree structure view of network topology and create
    *   segments of servers on that tree structure.
    *=========================================================================
    */

    static void doTraceroutes()
    {
	traceroutesStarted = true;
	// If segments are not built, build or read from disk
	if (assigned == null)
	  assigned = SerHashtable.factory("segments.ser");
	if (!assigned.isNew()) return;	// if it was deserialized, cancel segment building operations
	Runnable r = new SegmentBuilder();
	Thread t = new Thread(r);
	t.start();
	return;
    }


    /*
    *=========================================================================
    * doLatencies: Assigns each server to a separate segment by itself.
    *   Then, it sends dummy requests to other servers in order to
    *   measure their distances to the current server.
    *=========================================================================
    */

    static void doLatencies()
    {
	System.out.println("Latency measurement started");
	latenciesStarted = true;
	// If segments are not built, build or read from disk
	if (assigned == null)
	  assigned = SerHashtable.factory("segments.ser");
	if (!assigned.isNew()) return;	// if it was deserialized, cancel segment building operations

	// Segments were not found on disk, so build new segments. Assign each server
	//	to a segment by its own (to simplify, assign each segment to a segment
	//	numbered the same as the serverID number of that server).
	latenciesStarted = true;
	Enumeration enum = serverIDs.keys();
      String serverName = null;
	Integer segment = null, segSize;
	int seg = -1;
      while( enum.hasMoreElements() ) {
	  serverName = (String) enum.nextElement();
	  if (serverName == null) continue;
	  segment = new Integer((String) serverIDs.get(serverName));
	  if (segment == null) continue;
	  if (seg < (segment.intValue() + 1)) seg = segment.intValue() + 1;
	  assigned.put(serverName, segment);
	  segSize = new Integer(1);
        segmentSizes.put(segment, segSize);
	  System.out.println("Server " + serverName + " assigned to segment " + segment);
	}
	noOfSegments = seg;	// set Webpp global variable
	if (respTimes != null) {
	  respTimes.setNoOfSegments(seg);
	}
	assigned.serialize();

	// Start the latency distance measurement process
	//Runnable r = new LatencyMeasurer();
	//Thread t = new Thread(r);
	//t.start();
	return;
    }



    /*
    *=========================================================================
    * broadcastMessage: Sends a broadcast message to all alive servers
    *=========================================================================
    */

    static void broadcastMessage(String msgType, String param)
    {
	Runnable r = new MessageBroadcaster(msgType, param);
	Thread t = new Thread(r);
	t.start();
	return;
    }

   /*
    *=========================================================================
    * getSegmentofClient: finds out which segment the given client belongs to.
    *=========================================================================
    */
    static Integer getSegmentofClient(String client)
    {
      if (client == null || assigned == null) return null;
      String hostServer = (String) HostServer.get(client);
      if (hostServer == null) {
	System.out.println("Error: do not know the host server of " + client);
	return null;
      }
      if (assigned == null || !assigned.containsKey(hostServer)) {
	System.out.println("Error: do not know the segment of server " + hostServer);
	return null;
      }
      Integer seg = (Integer) assigned.get(hostServer);
      return seg;
    }

   /*
    *=========================================================================
    * insertRespTime: inserts the response time info into respTimes hash table
    *=========================================================================
    */
    static void insertRespTime(Integer clientSeg, String prevRespTimeStr)
    {
      int seg = clientSeg.intValue();
      int pos = prevRespTimeStr.indexOf('|');
      if (pos == -1) {
        System.out.println("Error: cannot separate fields of prevRespTimeStr");
        return;
      }
      String fileName = prevRespTimeStr.substring(0, pos);
      String tmp = prevRespTimeStr.substring(pos + 1);
      if (tmp == null) {
        System.out.println("Error: cannot get response time out of prevRespTimeStr");
        return;
      }
      long rtime = (new Long(tmp)).longValue();
      System.out.println("Inserting resp time: " + fileName + " seg: " +
			seg + " time: " + rtime);
      respTimes.put(fileName, seg, rtime);
      checkForReplication(fileName);
      if (totalRequestCount >= LimitToEnforceClearCount)
	respTimes.clear();	// Enforce a clear of all request counts
    }

    /*
    *=========================================================================
    * checkForReplication: performs a check to decide whether a new replica of the
    *	given resource is needed or not
    *=========================================================================
    */
    static void checkForReplication(String fname)
    {
	System.out.println("Check for Replication: " + fname);
	if (assigned == null) return;	// segments not built yet
	if (!respTimes.containsKey(fname)) return;
	if (!isHotResource(fname)) return;	// don't replicate if not a hot resource
	double maxPercentage = respTimes.getMostActiveSegmentPercentage(fname);
	if (maxPercentage < MinLoadPercentageForDecision)
	  return;	// don't replicate if there are no overloaded segments

	// Now decided to replicate. Get the segment number.
	Integer segment = respTimes.getMostActiveSegment(fname);
	System.out.println("I decided to replicate " + fname + " in segment " + segment);

	// Find an alive target server in the chosen segment
	String serverName = null;
	Vector candidateServers = new Vector();
	Enumeration enum = serverPorts.keys();
	while( enum.hasMoreElements() ) {
	  serverName = (String) enum.nextElement();
	  Integer seg = (Integer) assigned.get(serverName);
	  if (seg == null || !seg.equals(segment)) continue;	// skip if not in chosen segment
	  if (!serverIsAlive(serverName)) continue;	// skip if that server is dead
	  if (serverContainsResource(serverName, fname)) continue;	// skip if that server already has a replica
	  //serverName += ":" + ((String) serverPorts.get(serverName));
	  candidateServers.addElement(serverName);	// add server into candidates list
	}
	int chosenIndex;
	if (candidateServers.size() == 0) {
	  System.out.println("Error: cannot find a server for a new replica in segment " + segment);
	  return;
	}
	else if (candidateServers.size() == 1)
	  chosenIndex = 0;
	else {
	  // More than one candidate servers, choose one randomly
	  Random rnd = new Random(System.currentTimeMillis());
	  chosenIndex = rnd.nextInt() % candidateServers.size();
	}

	String chosenServer = (String) candidateServers.elementAt(chosenIndex);
	if (chosenServer == null) {
	  System.out.println("Error: chosenServer is null");
	  return;
	}

	System.out.println("No of Candidate Servers = " + candidateServers.size());
	System.out.println("ChosenIndex = " + chosenIndex);
	System.out.println("Chosen server for replication: " + chosenServer);

	// Perform replication, update replication directory, broadcast the directory.
	// First, get the target server's port number.
	String targetPort = (String) serverPorts.get(chosenServer);
	if (targetPort == null) {
	  System.out.println("Error: cannot figure out the port number of " + chosenServer);
	  return;
	}

	// Get the URL of local replica so that target server can request it from current server.
	String localURL = repl.getLocalURL(fname);
	if (localURL == null) {
	  System.out.println("Error: cannot figure out the local URL of " + fname);
	  return;
	}

	// Send the replication request to the target server, and leave the rest to it.
	try {
	  postURL(chosenServer + ":" + targetPort, "ReplicationRequest", fname + " " + localURL);
	} catch (Exception e) {
	  System.out.println("Error while sending ReplicationRequest to " + chosenServer);
	  return;
	}

	// Clear counts to restart with fresh request counts
	respTimes.clear();
    }

    /*
    *=========================================================================
    * checkForReplication: performs a check to decide whether a new replica of the
    *	given resource is needed or not
    *=========================================================================
    */
    static boolean isHotResource(String fname)
    {
	boolean answer = true;
	long count = respTimes.getRequestCount(fname);

	if (count < MinReqCountForDecision)
	  return false;		// don't replicate without enough information
	
	if (count < totalRequestCount * MinReqPercentageForHotDoc)
	  return false;

	System.out.println("Hot file: " + fname);
	return answer;
    }

    /*
    *=========================================================================
    * serverContainsResource: checks the replication directory to see if the
    *	given server already contains a replica of the given resource
    *=========================================================================
    */
    static boolean serverContainsResource(String serverName, String fname)
    {
	String replList = ReplicationDir.getUrls(fname);
	if (replList.indexOf(serverName) != -1) {
	  System.out.println("Server " + serverName + " already has a replica of " + fname);
	  return true;
	}
	return false;
    }

    /*
    *=========================================================================
    * postURL: send a POST request to the given server
    *=========================================================================
    */
    static void postURL(String serverName, String msgType, String param) throws Exception
    {
	byte b = 0;
	int done = 0, bytesRead;
	String urlName = "http://" + serverName + "/servlet/webppserv.Webpp/?" + msgType;
	URL url = new URL(urlName);
	HttpURLConnection conn = (HttpURLConnection) url.openConnection();
    
	conn.setRequestMethod("POST");
	bytesRead = 0;

	// Send your Post message
	if (param != null)
	  conn.setRequestProperty("param", param);

	// Get the response
	DataInputStream in = new DataInputStream(conn.getInputStream());
	try { b = in.readByte(); } catch (EOFException e) { done = 1; }
	while (done == 0) {
	  bytesRead++;
	  try { b = in.readByte(); } catch (EOFException e) { done = 1; }
	}
	in.close();
    }

    /*
    *=========================================================================
    * downloadResource: download a copy of the given resource from the given server
    *	at the given URL
    *=========================================================================
    */
    static void downloadResource(String remoteServer, String fname, String url)
    {
	Runnable r = new ResourceDownloader(remoteServer, fname, url);
	Thread t = new Thread(r);
	t.start();
	return;
    }

    /*
    *=========================================================================
    * getNoOfSegmentsFromTable: finds out the largest segment number in the
    *	'assigned' table. The no of segments is (largest seg no + 1). So, this
    *	function returns (largest seg no + 1) as the no of segments.
    *=========================================================================
    */
    static int getNoOfSegmentsFromTable()
    {
	if (assigned == null) return 0;
	int maxSeg = -1, seg;
	Enumeration enum = assigned.elements();
	while( enum.hasMoreElements() ) {
	  Integer sno = (Integer) enum.nextElement();
	  seg = sno.intValue();
	  if (seg > maxSeg) maxSeg = seg;
	}
	if (maxSeg == -1) {
	  System.out.println("Error: cannot determine the noOfSegments from the table");
	  return 0;
	}
	return maxSeg + 1;
    }

}	// class ends here
